#define STRICT
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void usage();
void cmdline(int argc, char* argv[]);
void readbmp(FILE* fp);
void writeicon(FILE* fp)   { fprintf(stderr, "### NOT IMPLEMENTED YET\n"); exit(1); }
void writesprite(FILE* fp) { fprintf(stderr, "### NOT IMPLEMENTED YET\n"); exit(1); }
void writepbmp(FILE* fp, int bpp, int mask);

/****************************************************************************
 *
 ****************************************************************************/

/* /usr/piece/include/piece_bmp.h */
typedef struct _PBMP_FILEHEADER {
	unsigned long head;		/* + 0,4 VOl`'PBMP' */
	unsigned long fsize;		/* + 4,4 t@CTCY   */
	unsigned char bpp;		/* + 8,1 rbg[x       */
	unsigned char mask;		/* + 9,1 }XN[x       */
	short w;			/* +10,2                */
	short h;			/* +12,2              */
					/* +14,2 pfBO       */
	unsigned long buf_size;		/* +16,4 f[^TCY     */
					/* =20                    */
} PBMP_FILEHEADER;

/****************************************************************************
 *	
 ****************************************************************************/

#define NULDEV	"NUL"
FILE* msg = stdout;

#define TYPE_TEXT	0
#define TYPE_BINARY	1
char* out_type_desc[] = {
	"TEXT",
	"BINARY",
};
int out_type = -1;

#define FMT_ICON	0
#define FMT_SPRITE	1
#define FMT_1BIT	2
#define FMT_1BITMASK	3
#define FMT_2BIT	4
#define FMT_2BITMASK	5
char* out_fmt_desc[] = {
	"ICON",
	"SPRITE",
	"1BIT",
	"1BIT MASK",
	"2BIT",
	"2BIT MASK",
};
int out_fmt = -1;

char out_label[32];
char out_prefix[32];

char in_path[_MAX_PATH];
char out_path[_MAX_PATH];
char drive[_MAX_DRIVE];
char dir[_MAX_DIR];
char fname[_MAX_FNAME];
char ext[_MAX_EXT];

#define MAX_SIZE	1024
unsigned char in_data[MAX_SIZE][MAX_SIZE];
int in_width;
int in_height;
int out_width;
int out_height;

/****************************************************************************
 *	main
 ****************************************************************************/

int
main(int argc, char* argv[])
{
	FILE* fp;

	/* R}hC͂܂B */
	cmdline(argc, argv);

	/* rbg}bvǂݍ݂܂B */
	fp = fopen(in_path, "rb");
	if(fp == NULL) {
		fprintf(stderr, "### %s open error\n", in_path);
		exit(1);
	}
	readbmp(fp);
	fclose(fp);

	/* o̓t@C쐬܂B */
	switch(out_type) {
	case TYPE_TEXT:
		fp = fopen(out_path, "wt");
		break;
	case TYPE_BINARY:
		fp = fopen(out_path, "wb");
		break;
	default:
		abort();
	}
	if(fp == NULL) {
		fprintf(stderr, "### %s create error\n", out_path);
		exit(1);
	}

	/* f[^o͂܂B */
	switch(out_fmt) {
	case FMT_ICON:
		writeicon(fp);
		break;
	case FMT_SPRITE:
		writesprite(fp);
		break;
	case FMT_1BIT:
		writepbmp(fp, 1, 0);
		break;
	case FMT_1BITMASK:
		writepbmp(fp, 1, 1);
		break;
	case FMT_2BIT:
		writepbmp(fp, 2, 0);
		break;
	case FMT_2BITMASK:
		writepbmp(fp, 2, 1);
		break;
	default:
		abort();
	}

	/* o̓t@C܂B */
	fclose(fp);

	return 0;
}

/****************************************************************************
 *	usage
 ****************************************************************************/

void
usage()
{
	fprintf(stderr, "USAGE:\n");
	fprintf(stderr, "	dpbmpcnv [options] <filename.bmp>\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "TYPE OPTIONS:\n");
	fprintf(stderr, "	-t         text                * default\n");
	fprintf(stderr, "	-b         binary\n");
	fprintf(stderr, "FORMAT OPTIONS:\n");
	fprintf(stderr, "	-fi        icon                  not implemented yet\n");
	fprintf(stderr, "	-fs        sprite                not implemented yet\n");
	fprintf(stderr, "	-f1        1bit data\n");
	fprintf(stderr, "	-f1m       1bit data with mask   extended\n");
	fprintf(stderr, "	-f2        2bit data\n");
	fprintf(stderr, "	-f2m       2bit data with mask * default\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "OTHER OPTIONS:\n");
	fprintf(stderr, "	-o<fname>  output filename\n");
	fprintf(stderr, "	-l<label>  label name\n");
	fprintf(stderr, "	-p<prefix> label prefix\n");
	fprintf(stderr, "	-q         quiet\n");
	exit(1);
}

/****************************************************************************
 *	cmdline
 ****************************************************************************/

void
cmdline(int argc, char* argv[])
{
	int i;
	char* opt;

	for(i = 1; i < argc; i++) {
		opt = argv[i];
		if(opt[0] != '-') {
			if(strlen(in_path) != 0) usage();
			strcpy(in_path, opt);
		} else {
			switch(opt[1]) {
			case 't':
				if(out_type != -1) usage();
				out_type = TYPE_TEXT;
				break;
			case 'b':
				if(out_type != -1) usage();
				out_type = TYPE_BINARY;
				break;
			case 'f':
				if(out_fmt != -1) usage();
				       if(strcmp(&opt[2], "i") == 0) {
					out_fmt = FMT_ICON;
				} else if(strcmp(&opt[2], "s") == 0) {
					out_fmt = FMT_SPRITE;
				} else if(strcmp(&opt[2], "1") == 0) {
					out_fmt = FMT_1BIT;
				} else if(strcmp(&opt[2], "1m") == 0) {
					out_fmt = FMT_1BITMASK;
				} else if(strcmp(&opt[2], "2") == 0) {
					out_fmt = FMT_2BIT;
				} else if(strcmp(&opt[2], "2m") == 0) {
					out_fmt = FMT_2BITMASK;
				} else {
					usage();
				}
				break;
			case 'o':
				if(strlen(out_path) != 0) usage();
				strcpy(out_path, &opt[2]);
				break;
			case 'l':
				if(strlen(out_label) != 0) usage();
				strcpy(out_label, &opt[2]);
				break;
			case 'p':
				if(strlen(out_prefix) != 0) usage();
				strcpy(out_prefix, &opt[2]);
				break;
			case 'q':
				msg = fopen(NULDEV, "wt");
				break;
			default:
				usage();
			}
		}
	}

	/* ^CvAtH[}bgw肳ĂȂ΁AftHglݒ肵܂B */
	if(out_type == -1) out_type = TYPE_TEXT;
	if(out_fmt == -1) out_fmt = FMT_2BITMASK;

	/* ̓t@C͕K{łB */
	if(strlen(in_path) == 0) usage();

	/* ̓t@C̊gqȗĂAu.bmpvƂ܂B */
	_splitpath(in_path, drive, dir, fname, ext);
	if(strlen(ext) == 0) {
		_makepath(in_path, drive, dir, fname, ".bmp");
	}

	/* o̓t@CȗĂA̓t@C琶܂B */
	if(strlen(out_path) == 0) {
		_splitpath(in_path, drive, dir, fname, ext);
		switch(out_type) {
		case TYPE_TEXT:
			_makepath(out_path, drive, dir, fname, ".c");
			break;
		case TYPE_BINARY:
			_makepath(out_path, drive, dir, fname, ".pgd");
			break;
		default:
			abort();
		}
	}

	/* o̓xȗĂA̓t@C琶܂B */
	if(strlen(out_label) == 0) {
		_splitpath(out_path, drive, dir, fname, ext);
		for(i = 0; fname[i] != '\0'; i++) {
			out_label[i] = toupper(fname[i]);
		}
	}

	/* o̓vtBNXw肳ĂAxɒǉ܂B */
	if(strlen(out_prefix) != 0) {
		strcat(out_prefix, out_label);
		strcpy(out_label, out_prefix);
	}

	fprintf(msg, "IN  FILE : %s\n", in_path);
	fprintf(msg, "OUT FILE : %s\n", out_path);
	fprintf(msg, "OUT TYPE : %s\n", out_type_desc[out_type]);
	fprintf(msg, "OUT FMT  : %s\n", out_fmt_desc[out_fmt]);
	fprintf(msg, "OUT LABEL: %s\n", out_label);
}

/****************************************************************************
 *	readbmp
 ****************************************************************************/

void
readbmp(FILE* fp)
{
	int retval, pitch, y;
	BITMAPFILEHEADER bf;
	BITMAPINFOHEADER bi;

	/* t@Cwb_ǂ݂܂B */
	retval = fread(&bf, 1, sizeof(BITMAPFILEHEADER), fp);
	if(retval != sizeof(BITMAPFILEHEADER)) goto ERR;
	if(strncmp((char*)&bf.bfType, "BM", 2) != 0) goto ERR;

	/* rbg}bvwb_ǂ݂܂B */
	retval = fread(&bi, 1, sizeof(BITMAPINFOHEADER), fp);
	if(retval != sizeof(BITMAPINFOHEADER)) goto ERR;
	if(bi.biSize < sizeof(BITMAPINFOHEADER)) goto ERR;
	if(bi.biWidth > MAX_SIZE) goto ERR;
	if(bi.biHeight > MAX_SIZE) goto ERR;
	if(bi.biPlanes != 1) goto ERR;   /* 256FpbgBMP̂ݑΉ */
	if(bi.biBitCount != 8) goto ERR; /* 256FpbgBMP̂ݑΉ */
	if(bi.biCompression != BI_RGB) goto ERR;

	in_width = bi.biWidth;
	in_height = bi.biHeight;
	pitch = in_width + 3 & ~3; /* DIB̎dl */

	retval = fseek(fp, bf.bfOffBits, SEEK_SET);
	if(retval != 0) goto ERR;
	for(y = in_height - 1; y >= 0; y--) { /* {gAbv */
		retval = fread(&in_data[y][0], 1, pitch, fp);
		if(retval != pitch) goto ERR;
	}

	out_width = in_width + 7 & ~7; /* PBMP̎dl */
	out_height = in_height;

	return;
ERR:
	fprintf(stderr, "### readbmp error\n");
	exit(1);
}

/****************************************************************************
 *	writebmp
 ****************************************************************************/

void
writepbmp(FILE* fp, int bpp, int mask)
{
	PBMP_FILEHEADER hdr;
	int x, y, i;
	unsigned char in_byte, out_byte;

	/* Cz̊JnB */
	if(out_type == TYPE_TEXT) {
		fprintf(fp, "const unsigned char %s[] = {\n", out_label);
	}

	/* wb_o͂܂B */
	memset(&hdr, 0, sizeof hdr);
	memcpy(&hdr.head, "PMBP", 4); /* VOl` */
	         hdr.fsize = sizeof hdr + out_width * out_height / (8 / bpp); /* wb_ƃJ[̃TCY */
	if(mask) hdr.fsize +=             out_width * out_height /  8;        /* }XÑTCYǉ   */
	hdr.bpp  = bpp;
	hdr.mask = mask;
	hdr.w = out_width;
	hdr.h = out_height;
	         hdr.buf_size =  out_width * out_height / (8 / bpp); /* J[̃TCY       */
	if(mask) hdr.buf_size += out_width * out_height /  8;        /* }XÑTCYǉ */
	if(out_type == TYPE_TEXT) {
		for(i = 0; i < sizeof hdr; i++) {
			fprintf(fp, "0x%02x,", ((unsigned char*)&hdr)[i]);
		}
		fprintf(fp, "\n");
	} else {
		fwrite(&hdr, 1, sizeof hdr, fp);
	}

	/* J[f[^o͂܂B */
	if(out_type == TYPE_TEXT) {
		fprintf(fp, "/*buf*/\n");
	}
	for(y = 0; y < out_height; y++) {
		for(x = 0; x < out_width; x += (8 / bpp)) {
			out_byte = 0;
			for(i = 0; i < (8 / bpp); i++) {
				out_byte <<= bpp;
				in_byte = in_data[y][x + i];
				if(in_byte < 4) {
					if(bpp == 1 && in_byte != 0) in_byte = 1; /* 1bit́AF1`31ƌȂ */
					out_byte |= in_byte;
				}
			}
			if(out_type == TYPE_TEXT) {
				fprintf(fp, "0x%02x,", out_byte);
			} else {
				fwrite(&out_byte, 1, 1, fp);
			}
		}
		if(out_type == TYPE_TEXT) {
			fprintf(fp, "\n");
		}
	}

	/* }XNf[^o͂܂B */
	if(mask) {
		if(out_type == TYPE_TEXT) {
			fprintf(fp, "/*mask*/\n");
		}
		for(y = 0; y < out_height; y++) {
			for(x = 0; x < out_width; x += 8) {
				out_byte = 0;
				for(i = 0; i < 8; i++) {
					out_byte <<= 1;
					in_byte = in_data[y][x + i];
					if(in_byte < 4) {
						out_byte |= 1;
					}
				}
				if(out_type == TYPE_TEXT) {
					fprintf(fp, "0x%02x,", out_byte);
				} else {
					fwrite(&out_byte, 1, 1, fp);
				}
			}
			if(out_type == TYPE_TEXT) {
				fprintf(fp, "\n");
			}
		}
	}

	/* Cz̏IB */
	if(out_type == TYPE_TEXT) {
		fprintf(fp, "};\n");
	}
}
